package org.tonylin.util.exec;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.net.SocketException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;

import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.easymock.PowerMock;
import org.powermock.modules.junit4.PowerMockRunner;

@RunWith(PowerMockRunner.class)
public class RetryableTaskTest {

	@Before
	public void setup(){
		PowerMock.resetAll();
	}
	
	@Test
	public void testVariableSleepStrategy(){
		ISleepStrategy sleepStrategy = new VariableSleepStrategy(new long[]{
				500, 1000, 2000
		});
		assertEquals(500, sleepStrategy.getSleepTime());
		assertEquals(1000, sleepStrategy.getSleepTime());
		assertEquals(2000, sleepStrategy.getSleepTime());
		try {
			 sleepStrategy.getSleepTime();
			 fail();
		} catch( Exception e ){
			//pass
		}
	}

	@Test
	public void testCallWithAlternativeMethod(){
		IUser user_expect = PowerMock.createMock(IUser.class);
		
		String errorMsg = "Testing error msg";
		final IUserDao userDao1 = PowerMock.createStrictMock(IUserDao.class);
		userDao1.getUser(EasyMock.anyObject(String.class));
		PowerMock.expectLastCall().andThrow(new RuntimeException(errorMsg)).once();
		
		final IUserDao userDao2 = PowerMock.createStrictMock(IUserDao.class);
		userDao2.getUser(EasyMock.anyObject(String.class));
		PowerMock.expectLastCall().andReturn(user_expect).once();
		
		PowerMock.replayAll();
		
		Callable<IUser> platformUtil = new Callable<IUser>() {
			private boolean switchFlag = true;
			@Override
			public IUser call() throws Exception {
				switchFlag = !switchFlag;
				if( !switchFlag )
					return userDao1.getUser("1234");
				else {
					return userDao2.getUser("1234");
				}
			}
		};
		
		IRetryableTask<IUser> retryableTask = new RetryableTask<IUser>(platformUtil);
		IRetryablePolicy compositeRetryablePolicy = new CompositeRetryablePolicy(Arrays.asList(new IRetryablePolicy[]{
				new AttemptRetryablePolicy(3),
				new ExceptionRetryablePolicy()
		}));
		retryableTask.setRetryablePolicy(compositeRetryablePolicy);
		try {
			assertEquals(user_expect, retryableTask.call());
		} catch (Throwable e) {
			fail();
		}
		
		PowerMock.verifyAll();
	}
	
	@Test
	public void testCallWithNullRetryablePolicy(){
		final IUser user_expect = PowerMock.createMock(IUser.class);
		final IUserDao userDao = PowerMock.createStrictMock(IUserDao.class);
		userDao.getUser("1234");
		PowerMock.expectLastCall().andAnswer(new IAnswer<IUser>() {
			private int times = 0;
			@Override
			public IUser answer() throws Throwable {
				if( times++ == 0 ){
					return null;
				} else {
					return user_expect;
				}
			}
		}).times(2);
		
		PowerMock.replayAll();
		
		Callable<IUser> platformUtil = new Callable<IUser>() {
			@Override
			public IUser call() throws Exception {
				return userDao.getUser("1234");
			}
		};
		
		IRetryableTask<IUser> retryableTask = new RetryableTask<IUser>(platformUtil);
		IRetryablePolicy compositeRetryablePolicy = new CompositeRetryablePolicy(Arrays.asList(new IRetryablePolicy[]{
				new AttemptRetryablePolicy(3),
				new NullRetryablePolicy()
		}));
		retryableTask.setRetryablePolicy(compositeRetryablePolicy);
		try {
			assertEquals(user_expect, retryableTask.call());
		} catch (Throwable e) {
			fail();
		}
		PowerMock.verifyAll();
	}
	
	@Test
	public void testCallWithExceptionRetryablePolicy(){
		final String errorMsg = "Testing error msg";
		
		Callable<IUser> platformUtil = new Callable<IUser>() {
			private int count = 0;
			@Override
			public IUser call() throws Exception {
				switch (count++) {
				case 0:
					throw new SocketException();
				case 1:
					throw new RemoteException();
				default:
					throw new RuntimeException(errorMsg);
				}
			}
		};

		List<Class<? extends Throwable>> exceptionList = new ArrayList<Class<? extends Throwable>>();
		exceptionList.add(SocketException.class);
		exceptionList.add(RemoteException.class);
		
		IRetryableTask<IUser> retryableTask = new RetryableTask<IUser>(platformUtil);
		IRetryablePolicy compositeRetryablePolicy = new CompositeRetryablePolicy(Arrays.asList(new IRetryablePolicy[]{
				new AttemptRetryablePolicy(3),
				new ExceptionRetryablePolicy(exceptionList)
		}));
		retryableTask.setRetryablePolicy(compositeRetryablePolicy);
		try {
			retryableTask.call();
			fail();
		} catch (Throwable e) {
			assertEquals(errorMsg, e.getMessage());
		}
	}		
	
	@Test
	public void testCallWithAttemptRetryablePolicy(){
		String errorMsg = "Testing error msg";
		final IUserDao userDao = PowerMock.createStrictMock(IUserDao.class);
		userDao.getUser(EasyMock.anyObject(String.class));
		PowerMock.expectLastCall().andThrow(new RuntimeException(errorMsg)).times(6);
		
		PowerMock.replayAll();
		
		Callable<IUser> platformUtil = new Callable<IUser>() {
			@Override
			public IUser call() throws Exception {
				return userDao.getUser("1234");
			}
		};
		
		IRetryableTask<IUser> retryableTask = new RetryableTask<IUser>(platformUtil);
		retryableTask.setRetryablePolicy(new AttemptRetryablePolicy(5));
		try {
			retryableTask.call();
			fail();
		} catch (Throwable e) {
			assertEquals(errorMsg, e.getMessage());
		}
		
		PowerMock.verifyAll();
	}	
	
	@Test
	public void testCall_normal(){
		final String uid = "1234";
		IUser user_expect = PowerMock.createMock(IUser.class);
		
		final IUserDao userDao = PowerMock.createMock(IUserDao.class);
		userDao.getUser(uid);
		PowerMock.expectLastCall().andReturn(user_expect).once();
		
		PowerMock.replayAll();
		
		Callable<IUser> platformUtil = new Callable<IUser>() {
			@Override
			public IUser call() throws Exception {
				return userDao.getUser(uid);
			}
		};
		
		IRetryableTask<IUser> retryableTask = new RetryableTask<IUser>(platformUtil);
		try {
			IUser user_real = retryableTask.call();
			assertEquals(user_expect, user_real);
		} catch (Throwable e) {
			fail();
		}
	}
	
}
